// record-mmap.c
// Simple Adress-Datenbank, Variante mit mmap()
// Hans-Georg Esser, 07.05.2013

#include <string.h>  // strcpy, strcmp
#include <fcntl.h>   // open
#include <stdio.h>   // printf, scanf
#include <unistd.h>  // lseek
#include <sys/stat.h>    // S_IWUSR etc. 
#include <sys/mman.h>    // fuer mmap
 
typedef struct {
  int  id : 15;      // 15 Bits
  int  deleted : 1;  // 1 Bit
  char name[30];
  char vorname[30];
  char telefon[30];
  char gebdatum[11];
} record;

int recsize = sizeof(record);

record *database;    // Darueber sprechen wir die DB im Speicher an
record *database_start;

int main () {
  record r;
  int fd;
  char filename[]="adressen.dat";
  char action[100];
  
  strncpy (r.name, "Mueller", 29);
  strncpy (r.vorname, "Anton", 29);
  strncpy (r.telefon, "0172/12345", 29);
  strncpy (r.gebdatum, "25.12.1980", 10);
  r.deleted = 0;

  int i;
  fd = open (filename, O_RDWR | O_CREAT, S_IRUSR | S_IWUSR);

  if ( lseek(fd,0,SEEK_END) == 0 ) {
    // Datei ist leer; Standardinhalte erzeugen
    printf ("open: Initialisiere Daten\n");
    for (i=0; i<100; i++) {
      r.id = i;
      write (fd, &r, recsize);
    }
  } else printf ("open: Daten existieren schon\n");


  // ab hier Aenderungen (mit mmap)
  database_start = mmap (0, 100*sizeof(record), PROT_READ | PROT_WRITE,
    MAP_SHARED, fd, 0);
  close (fd);   // brauchen offene Datei nach Mapping nicht mehr
  database = database_start;

  printf ("Aktionen: show n, edit n, del n, new 0, search 0, quit 0\n");

  while (1) {
    printf ("Aktion Nummer: ");
    scanf ("%s %d", (char*)&action, &i);

    // action: show
    if (!strcmp((char*)&action, "show")) {
      // lseek (fd, i*recsize, SEEK_SET);
      // read (fd, &r, recsize);
      r = database[i];
      if (!r.deleted) {
        printf ("ID:   %d\n", r.id);
        printf ("Name: %s %s\n", r.vorname, r.name);
        printf ("Tel.: %s\n", r.telefon);
        printf ("Geb.: %s\n", r.gebdatum);
      } else printf ("Record %d deleted\n", i);
    };

    // action: edit
    if (!strcmp((char*)&action, "edit")) {
      // lseek (fd, i*recsize, SEEK_SET);
      // read (fd, &r, recsize);
      r = database[i];
      if (!r.deleted) { 
        printf ("ID: %d, Name: %s %s\n", r.id, r.vorname, r.name);
        printf ("Eingabe: Name Vorname Telefon Geburtstag: ");
        scanf ("%s %s %s %s", r.name, r.vorname, r.telefon, r.gebdatum);
        // lseek (fd, i*recsize, SEEK_SET);
        // write (fd, &r, recsize);
        database[i] = r;
      } else printf ("Record %d deleted\n", i);
    };
    
    // action: delete
    if (!strcmp((char*)&action, "del")) {
      // lseek (fd, i*recsize, SEEK_SET);
      // read (fd, &r, recsize);
      r = database[i];
      r.deleted = 1;
      // lseek (fd, i*recsize, SEEK_SET);
      // write (fd, &r, recsize);
      database[i] = r;
    };

    // action: new (Aufgabe 7 b)
    if (!strcmp((char*)&action, "new")) {
      printf ("Nicht implementiert.\n");
      /*  
        Das geht mit mmap nicht so einfach
        
      lseek (fd, -recsize, SEEK_END);  // auf letzten Eintrag springen
      read (fd, &r, recsize);
      r.id++;                          // ID um 1 erhhen
      r.deleted = 0;                   // nicht gelscht
      printf ("Eingabe: Name Vorname Telefon Geburtstag: ");
      scanf ("%s %s %s %s", r.name, r.vorname, r.telefon, r.gebdatum);
      write (fd, &r, recsize);
      */
    };
    
    // action: search (Aufgabe 7 c)
    if (!strcmp((char*)&action, "search")) {
      char search[100];
      int i;
      printf ("Suchbegriff: ");
      scanf ("%s", (char*)&search);
      // lseek (fd, 0, SEEK_SET);
      // while (read (fd, &r, recsize)) {
      for (i=0; i<100; i++) {
        r = database[i];
        if ((r.deleted == 0) && 
            ((strcmp(r.vorname, search) == 0) ||
             (strcmp(r.name, search) == 0)))
          printf ("Treffer in record %d, %s %s\n", r.id, r.vorname, r.name);
      }
    }

    // action: quit
    if (!strcmp((char*)&action, "quit")) {
      munmap (database, 100*sizeof(record));
      printf ("Programmende.\n");
      break;
    };

  };

};
